
/*
   helmath1.c
   
   Math routines for helicon.
   
   Author: Mark E. Nash
   Date:   June, 1988
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<malloc.h>
#include	<math.h>
#include	<string.h>
#include	<conio.h>
#include	"helimacs.h"
#include	"helimess.h"
#include	"helstruc.h"
#include	"helfdef.h"


/* externals */
extern char msg[];				/* in heliglob.h */
extern char *fitlist[];				/* in heliglob.h */
extern struct rawcell *rd_top;			/* in heliglob.h */
extern struct menutop *menus;			/* in heliglob.h */
extern double params[MAXPARS+1][NUMDIST+1];	/* in heliglob.h */
extern int lastnum;				/* in heliglob.h */


int makebins(traw, tbins, num, class)
	struct rawcell *traw;
	struct topbin **tbins;
	int *num;
	int class;
/*
   makebins() generates the frequency distribution of the raw data.
   
   Entry: traw points to the top of the raw data list.
	  num is the number of bins to generate.
	  class is the distribution class = DISCRETE or CONTINUOUS.

   Exit:  If no errors, the bin list has been generated, and *tbins points to
   		the top of the list.
	  If errors, returns ERR and *tbins = NULL,
	  else NOERR.
*/
	{
	struct topbin *tb;		/* used for initial allocation */
	struct bin *bp;			/* bin pointer */
	struct rawcell *rd;		/* step through the raw data list */
	unsigned binctr;		/* count the bins */
	double	chkval;			/* to see if raw data is in order */
	double	datval;			/* raw data value */
	int	ctr;			/* gp counter */
	struct grafwindow *ms;		/* message window */
	struct menu *mp;		/* to stroll through the menu list */
	struct menuitem *ip;		/* to stroll through the item list */


/* display calculation message */
	if ((ms = msg_window(I_BCALC)) == NULL)
		{
		printf("\7\7\7");
		return(ERR);
		}
		
/* see if old parameters are still valid */
	if (*num != lastnum)
		{
/* reset FITQ flags and error values when bins are re-made */
		for (ctr=NULLDIST+1; ctr<=NUMDIST; ctr++)
			{
			params[FITQ][ctr] = DOFIT;
			params[MSE][ctr] = -1.0;
			}

/* reset fit menu hilights */
		mp = menus->start;
		while (mp->menuid != FITID)
			mp = mp->next;

		ip = mp->top;
		while (ip != NULL)
			{
			ip->hf = FALSE;
			ip = ip->next;
			}
		}
		
	lastnum = *num;

/* check for bad data */
	if (*num <= 0)
		{
		err_window(M_NOBINS);
		return(ERR);
		}
		
	*tbins = NULL;
/* allocate the top of the bin list */
	if ((tb = (struct topbin *)malloc(sizeof(struct topbin))) == NULL)
		{
		printf("\7");
		return(ERR);
		}
/* update memory status */
	memupdt((unsigned long)sizeof(struct topbin), ALLOC);

/* set the class */
	tb->dclass = class;
	
/* get extreme x values.  Since the raw data is in ascending x order, get
   the first and last x values. */

/* get the min value */
   	rd = traw;
	if (rd != NULL)
		{
		if (class == CONTINUOUS)
			tb->xmin = rd->x;
		else
			tb->xmin = (double)((int)(rd->x + 0.5));

		chkval = rd->x;
		}
	else
		{
		err_window(D_NOMINX);
		free_bins(&tb);
		return(ERR);
		}
		
/* now get the max value */
	while (rd->next != NULL)
		{
		rd = rd->next;
		if (chkval > rd->x)
			{
			sprintf(msg, D_ORDER, chkval, rd->x);
			err_window(msg);
			}
		chkval = rd->x;
		}

	if (class == CONTINUOUS)
		tb->xmax = rd->x;
	else
		tb->xmax = (double)((int)(rd->x + 0.5));
		
/* set the bin size */
	if (class == CONTINUOUS)
		{
		tb->binsize = (tb->xmax - tb->xmin) / (double)(*num);
		tb->dbinsize = 0;
		}
	else
		{
/* for discrete case, set the number of bins */
		*num = (int)fabs(tb->xmax - tb->xmin) + 1;
		tb->binsize = 0.0;
		tb->dbinsize = 1;
		}
	
/* null out the pointer */
	tb->start = NULL;
	
/* get the mean and variance of the raw data */
	meancalc(rd_top, &(tb->mean));
	varcalc(rd_top, tb->mean, &(tb->var));

/* allocate the bin list */
/* do the first one to save the start */
	if ((tb->start = (struct bin *)malloc(sizeof(struct bin))) == NULL)
		{
		printf("\7");
		return(ERR);
		}
/* update memory status */
	memupdt((unsigned long)sizeof(struct bin), ALLOC);

	binctr = 0;
	rd = traw;
	bp = tb->start;
	bp->next = NULL;
	bp->xval = tb->xmin;
	if (class == CONTINUOUS)
		{
		bp->x = tb->xmin + tb->binsize / 2.0;
		datval = rd->x;
		}
	else
		{
		bp->x = tb->xmin;
		datval = (double)((int)(rd->x + 0.5));
		}
		
	bp->y = 0.0;
	while (datval <= (bp->xval + tb->binsize))
		{
		bp->y += 1.0;
		rd = rd->next;
		if (rd == NULL)
			{
			sprintf(msg, D_RANOUT, tb->xmax, binctr, *num);
			err_window(msg);
			free_bins(&tb);
			return(ERR);
			}

		if (class == CONTINUOUS)
			datval = rd->x;
		else
			datval = (double)((int)(rd->x + 0.5));
		}

/* assign y extremes */
	tb->ymin = bp->y;
	tb->ymax = bp->y;
	
/* generate the rest of the bins, except the last one */
	for (binctr++; binctr<(*num)-1; binctr++)
		{
		if ((bp->next = (struct bin *)malloc(sizeof(struct bin)))
								== NULL)
			{
			printf("\7");
			return(ERR);
			}
/* update memory status */
		memupdt((unsigned long)sizeof(struct bin), ALLOC);

		bp = bp->next;
		bp->next = NULL;
		if (class == CONTINUOUS)
			{
			bp->xval = tb->xmin + tb->binsize * (double)binctr;
			bp->x = bp->xval + tb->binsize / 2.0;
			datval = rd->x;
			}
		else
			{
			bp->xval = tb->xmin + (double)binctr;
			bp->x = bp->xval;
			datval = (double)((int)(rd->x + 0.5));
			}
			
		bp->y = 0.0;
		while (datval <= (double)(bp->xval + tb->binsize))
			{
			bp->y += 1.0;
			rd = rd->next;
			if (rd == NULL)
				{
				sprintf(msg, D_RANOUT, tb->xmax, binctr, *num);
				err_window(msg);
				free_bins(&tb);
				return(ERR);
				}
				
			if (class == CONTINUOUS)
				datval = rd->x;
			else
				datval = (double)((int)(rd->x + 0.5));
			}

/* reassign y extremes if necessary */
		if (bp->y < tb->ymin)
			tb->ymin = bp->y;
		else
			if (bp->y > tb->ymax)
				tb->ymax = bp->y;
		}

/* now generate the last bin and put all the rest of the data in it.  This is
   done because the end point is a maxima that might not be tested correctly
   if used in the above loop.  There should always be at least 1 more point
   in the list at this time (which is the maxima), so there is no need to test
   rd for null at the start of this section */
   
	if ((bp->next = (struct bin *)malloc(sizeof(struct bin))) == NULL)
		{
		printf("\7");
		return(ERR);
		}
/* update memory status */
	memupdt((unsigned long)sizeof(struct bin), ALLOC);

	bp = bp->next;
	bp->next = NULL;
	if (class == CONTINUOUS)
		{
		bp->xval = tb->xmin + tb->binsize * (double)binctr;
		bp->x = bp->xval + tb->binsize / 2.0;
		datval = rd->x;
		}
	else
		{
		bp->xval = tb->xmin + (double)binctr;
		bp->x = bp->xval;
		datval = (double)((int)(rd->x + 0.5));
		}
			
	bp->y = 0.0;
	while (rd != NULL)
		{
		bp->y += 1.0;
		rd = rd->next;
		}

/* reassign y extremes if necessary */
	if (bp->y < tb->ymin)
		tb->ymin = bp->y;
	else
		if (bp->y > tb->ymax)
			tb->ymax = bp->y;
   

/* count the total number of points and save it */
	for (bp=tb->start, tb->numpts = 0.0; bp!=NULL; bp=bp->next)
		tb->numpts += bp->y;
		
/* attach list to top */
	*tbins = tb;
	
	close_grafwin(&ms);
	return(NOERR);
	}



int free_bins(top)
	struct topbin **top;
/*
   free_bins() liberates the bin list.
   
   Entry: *top points to the top of the list.
	  
   Exit:  *top points to NULL.
          If no errors, the list has been distroyed,
          else returns ERR.
*/
	{
	struct bin *bp, *temp;			/* bin pointers */
	
	
	if (*top == NULL)
		return(NOERR);

	bp = (*top)->start;
	while (bp != NULL)
		{
		temp = bp;
		bp = bp->next;
		free((void *)temp);
/* update memory status */
		memupdt((unsigned long)sizeof(struct bin), FRE);
		}
		
	free((void *)(*top));
/* update memory status */
	memupdt((unsigned long)sizeof(struct topbin), FRE);
	*top = NULL;

/* check the heap */
	switch (_heapchk())
		{
		case _HEAPOK:
		case _HEAPEMPTY:
			break;
		case _HEAPBADBEGIN:
		case _HEAPBADNODE:
			printf("\7");
			return(ERR);
		}

	return(NOERR);
	}
	
	
	
int tickspace(mingraph, maxgraph, start, space, minspace)
	double mingraph, maxgraph;
	double *start, *space, *minspace;
/*
   ticspace() calculates the spacing for tick marks (same as Micro Saint).

   Entry: mingraph, maxgraph = bounds of the graph
   
   Exit:  start = where to put the first major tick mark
	  space = spacing from one major tick mark to the next
	  minspace = spacing from one minor tick mark to the next
*/
	{
	double span;			/* span between min and max */
	double factor;				/* for scaling the span */
	double divisor;				/* for scaling the span */
	double temp;				/* temporary storage */
	double temp2;				/* more temp storage */
	int division[3];			/* allowed divisions */
	int minors[3];		/* number of minor divisions per major */
	int numticks[3];			/* number of tick marks */
	long firsttick[3];			/* where the first tick goes */
	int d;					/* array index */
	long place;			/* where a tick mark is placed */
	int best;				/* best number of tick marks */
	int picked;				/* which division is picked */

						/* initialize stuff */
	division[0] = 1;
	division[1] = 2;
	division[2] = 5;
	minors[0] = 5;
	minors[1] = 4;
	minors[2] = 5;

				/* We use both factor and divisor because
				   we can't store tenths accurately. */
	factor = 1.0;
	divisor = 1.0;

						/* find span of the graph */
	span = maxgraph - mingraph;

				/* scale it so that 3.0 < span <= 30.01 */
	if (span > 30.01)
		for (divisor = 10.0; span / divisor > 30.01; divisor *= 10.0);
	else if (span <= 30.01 && span > 3.0)
		{}
	else
	       for (factor = 10.0; factor * span - .01 <= 3.0; factor *= 10.0);

				/* check each division, and see how many tick
				   marks would result if it were used	*/
	for (d = 0; d < 3; d++)
		{
					/* find where the first tick goes */
		temp = mingraph * factor / divisor + division[d];
		modf(temp, &temp2);
		if (temp - temp2 == 0.0)
			temp -= .01;
				/* beware truncation of negative numbers ! */
		if (temp >= 0.0)
			firsttick[d] = (long)temp;
		else
			firsttick[d] = (long)(temp - 1.0);
		while (firsttick[d] % division[d] != 0)
			firsttick[d]--;

					/* count off places where ticks go */
		numticks[d] = 0;
		for (place = firsttick[d];
			place <= (long)(maxgraph * factor / divisor);
			place += (long)division[d])
			numticks[d]++;
		}

				/* pick the division which results in the
				   least number of tick marks >= 3 */
					/* put illegals outta the running */
	for (d = 0; d < 3; d++)
		{
		if (numticks[d] < 2)
			numticks[d] += 100;
		if (numticks[d] == 2)		/* 2 as a last resort */
			numticks[d] += 50;
		}
					/* get the best one */
	best = numticks[0];
	picked = 0;
	for (d = 1; d < 3; d++)
		if (numticks[d] < best)
			{
			picked = d;
			best = numticks[d];
			}

				/* calculate the final answers */
	*start = firsttick[picked] * divisor / factor;
	*space = division[picked] * divisor / factor;
	*minspace = *space / minors[picked];

	return(TRUE);
	}
	
	
	
int fmeancalc(top, retval)
	struct bin *top;
	double *retval;
/*
   fmeancalc() calculates the mean for a frequency distribution.
   
   Entry: top points to the first bin in the bin list.
   
   Exit:  If no errors, *retval = mean and returns NOERR.
          Else, returns ERR.
	  
   Note: mean = S(x*y) / S(y),  where S is the sum over all points.
*/
	{
	struct bin *bp;			/* step through bin list */
	double	num, denom;		/* for calculation */
	

	*retval = 0.0;
	num = 0.0;
	denom = 0.0;

	if (top == NULL)
		{
		err_window(M_NOMEAN);
		return(ERR);
		}
	
	bp = top;
	for (bp=top; bp!=NULL; bp=bp->next)
		{
		num += bp->x * bp->y;
		denom += bp->y;
		}
		
	*retval = num / denom;
	
	return(NOERR);
	}
	
	
	
int fvarcalc(top, mean, retval)
	struct bin *top;
	double mean;
	double *retval;
/*
   fvarcalc() calculates the variance for a frequency distribution.
   
   Entry: top points to the first bin in the bin list.
   
   Exit:  If no errors, *retval = variance and returns NOERR.
          Else, returns ERR.

   Note:  var = (S(x^2 * y) - n * mean^2) / (n-1) where S is the sum 
          over all pts.
*/
	{
	struct bin *bp;			/* step through bin list */
	double	num, denom;		/* for calculation */
	double	num2, s2;
	

	*retval = 0.0;
	num = 0.0;
	denom = 0.0;
	num2 = 0.0;

	if (top == NULL)
		{
		err_window(M_NOVAR);
		return(ERR);
		}
	
	bp = top;
	for (bp=top; bp!=NULL; bp=bp->next)
		{
		num2 += bp->y * pow((bp->x - mean), 2.0);
		num += pow(bp->x, 2.0) * bp->y;
		denom += bp->y;
		}
		
	s2 = num2 / (denom - 1);
	*retval = (num - denom * pow(mean, 2.0)) / (denom - 1);
		
	return(NOERR);
	}
	
	
	
int meancalc(top, retval)
	struct rawcell *top;
	double *retval;
/*
   meancalc() calculates the mean for a set of data.
   
   Entry: top points to the first raw data item in the list.
   
   Exit:  If no errors, *retval = mean and returns NOERR.
          Else, returns ERR.
	  
   Note: mean = S(x) / n, where S is the sum over x, 
         and n is the number of x's.
*/
	{
	struct rawcell *rp;		/* step through raw data list */
	double	num, denom;		/* for calculation */
	

	*retval = 0.0;
	num = 0.0;
	denom = 0.0;

	if (top == NULL)
		{
		err_window(M_NOMEAN);
		return(ERR);
		}
	
	rp = top;
	for (rp=top; rp!=NULL; rp=rp->next)
		{
		num += rp->x;
		denom += 1.0;
		}
		
	*retval = num / denom;
	
	return(NOERR);
	}
	
	
	
int varcalc(top, mean, retval)
	struct rawcell *top;
	double mean;
	double *retval;
/*
   varcalc() calculates the variance for a set of data.
   
   Entry: top points to the first data item in the raw data list.
          mean is the mean of the raw data.
   
   Exit:  If no errors, *retval = variance and returns NOERR.
          Else, returns ERR.

   Note:  var = (S(x^2) - n * mean^2) / (n - 1)
*/
	{
	struct rawcell *rp;		/* step through bin list */
	double	num, denom;		/* for calculation */
	

	*retval = 0.0;
	num = 0.0;
	denom = 0.0;

	if (top == NULL)
		{
		err_window(M_NOVAR);
		return(ERR);
		}
	
	rp = top;
	for (rp=top; rp!=NULL; rp=rp->next)
		{
		num += pow(rp->x, 2.0);
		denom += 1.0;
		}
		
	if (denom < 2.0)
		{
		err_window(M_NOVPT);
		return(ERR);
		}
		
	*retval = (num - denom * pow(mean, 2.0)) / (denom - 1);
		
	return(NOERR);
	}
	
	
	
int magcalc(top, dtype, pa, retval)
	struct bin *top;
	unsigned dtype;
	double pa[MAXPARS+1][NUMDIST+1];
	double *retval;
/*
   magcalc() calculates the magnitude for a set of data.
   
   Entry: top points to the first bin in the bin list.
          dtype is the distribution type.
	  pa is the parameter array
   
   Exit:  If no errors, *retval = magnitude and returns NOERR.
          Else, returns ERR.
*/
	{
	struct bin *bp;			/* step through bin list */
	double	num, denom;		/* numerator & denominator calc. */
	double	y;			/* y-value calc. */
	

	*retval = 0.0;
	
/* calculate the magnitude:  mag = [sum of (f(x)*y)] / sum of f(x)^2
   over all data points */
	
	if (top == NULL)
		{
		err_window(M_NOMAG);
		return(ERR);
		}
	
	bp = top;
	num = 0.0;
	denom = 0.0;
	for (bp=top; bp!=NULL; bp=bp->next)
		{
		if (fval(dtype, bp->x, pa, TRUE, &y))
			if (bp->y != 0.0)	/* leave out the null areas */
				{
				num += bp->y * y;
				denom += pow(y, 2.0);
				}
			else;
		else;		/* bad domain value */
		}
	
	if (denom == 0.0)
		{
		sprintf(msg, M_BADNOM, "magcalc()");
		err_window(msg);
		return(ERR);
		}

	*retval = num / denom;
	
	return(NOERR);
	}
	
	
	
double chitest(top, ftype, pa, np)
	struct bin *top;
	unsigned ftype;
	double pa[MAXPARS+1][NUMDIST+1];
	double np;
/*
   chitest() calculates the chi-square goodness-of-fit test.
   
   Entry: top points to the top of the bin list.
          ftype is the function type.
	  pa is the parameter array, which has been set.
	  np is the total number of data points.
	  
   Exit:  Returns a positive mse if no error,
          Else returns MSERR if error.
*/
	{
	struct bin *bp;			/* pointer to the data */
	double	num, denom;		/* for calculation */
	double	retval;			/* number to return */
	double	y;			/* for y calc */
	
	
	if (top == NULL)
		{
		err_window(M_NOCS);
		return(MSERR);
		}
		
	num = 0.0;
	denom = 0.0;
	for (bp=top; bp!=NULL; bp=bp->next)
		{
		if (fval(ftype, bp->x, pa, TRUE, &y))
			if (bp->y != 0.0)
				{
				num += pow(bp->y - y, 2.0);
				denom += y;
				}
			else;
		else;		/* bad domain value */
		}

	if (denom == 0.0)
		{
		sprintf(msg, M_BADNOM, "chitest()");
		err_window(msg);
		return(MSERR);
		}

	retval = num / denom;
		
	return(retval);
	}


	
double mse(top, ftype, pa)
	struct bin *top;
	unsigned ftype;
	double pa[MAXPARS+1][NUMDIST+1];
/*
   mse() calculates the mean squared error of a set of data.
   
   Entry: top points to the top of the bin list.
          ftype is the function type.
	  pa is the parameter array, which has been set.
	  
   Exit:  Returns a positive mse if no error,
          Else returns MSERR if error.

   Note:  mse = S((f(x) - y)^2) / n,  S is the sum over n points.
*/
	{
	struct bin *bp;			/* pointer to the data */
	double	num, denom;		/* for calculation */
	double	retval;			/* number to return */
	double	y;			/* for y calc */
	
	
	if (top == NULL)
		{
		err_window(M_NOMSE);
		return(MSERR);
		}
		
	num = 0.0;
	denom = 0.0;
	for (bp=top; bp!=NULL; bp=bp->next)
		{
		if (fval(ftype, bp->x, pa, FALSE, &y))
			if (bp->y != 0.0)
				{
				num += pow(y - bp->y, 2.0);
				denom += 1.0;
				}
			else;
		else;		/* bad domain */
		}

	if (denom == 0.0)
		{
		sprintf(msg, M_BADNOM, "mse()");
		err_window(msg);
		return(MSERR);
		}

	retval = num / denom;
		
	return(retval);
	}


	
int fval(dt, x, pa, norm, yval)
	unsigned dt;
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *yval;
/*
   fval() gets a value from a distribution.
   
   Entry: dt is the distribution type.
          x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm is TRUE for normalized calculation, otherwise FALSE.
	  
   Exit:  A point from the specified distribution is returned.
*/
	{
	switch (dt)
		{
		case RECT:
			return(rectgen(x, pa, norm, yval));
		case TRI:
			return(trigen(x, pa, norm, yval));
		case NOR:
			return(normgen(x, pa, norm, yval));
		case GAM:
			return(gamgen(x, pa, norm, yval));
		case PAR:
			return(pargen(x, pa, norm, yval));
		case FDIST:
			return(fgen(x, pa, norm, yval));
		case BETA:
			return(betagen(x, pa, norm, yval));
		case LOGNOR:
			return(lnorgen(x, pa, norm, yval));
		case EXPON:
			return(exgen(x, pa, norm, yval));
		case DEXPON:
			return(dexgen(x, pa, norm, yval));
		case GUM:
			return(gumgen(x, pa, norm, yval));
		case CHI:
			return(chigen(x, pa, norm, yval));
		case LOGIS:
			return(loggen(x, pa, norm, yval));
		case TDIST:
			return(tgen(x, pa, norm, yval));
		case WEI:
			return(weigen(x, pa, norm, yval));
		case BINOM:
			return(bingen(x, pa, norm, yval));
		case POI:
			return(poigen(x, pa, norm, yval));
		case GEO:
			return(geogen(x, pa, norm, yval));
		default:
			sprintf(msg, P_BADDT, dt);
			err_window(msg);
			return(FALSE);
		}
	}

	
	
double binco(n, x)
	long n, x;
/*
   binco() calculates the binomial coefficient of n choose x.
   
   Entry: n is the number in the group.
          x is the number to choose.
	   
   Exit:  returns the binomial coefficient.
*/
	{
	double coeff;
	double num, denom;
	

	if (x < 0)
		coeff = 0.0;
		
	else 
		if (x == 0)
			coeff = 1.0;
		else 
			if (x > (n-x))
				{
				num = topfact(n, x);
				denom = fact((int)(n-x));
				coeff = num / denom;
				}
			else
				{
				num = topfact(n, (n-x));
				denom = fact((int)x);
				coeff = num / denom;
				}
	return(coeff);
	}
	
	
	
double fact(x)
	int x;
/*
   fact() generates x!
   
   Entry: x is the argument for the function.
   
   Exit:  returns x!
*/
	{
	double	ctr, retval;		/* for calculations */
	
	
	retval = 1.0;
	if (x < 1)
		return(retval);
		
	for (ctr=(double)x; ctr>1.0; ctr-=1.0)
		retval *= ctr;
	
	return(retval);
	}

	
		
double topfact(n, x)
	long n, x;
/*
   topfact() generates n * (n-1) * (n-2) * ... * (x+2) * (x+1) * x.
   
   Entry: n and x are the arguments for the function.
   
   Exit:  returns n * (n-1) * (n-2) * ... * (x+2) * (x+1) * x.
*/
	{
	double	ctr, retval;		/* for calculations */
	
	
	retval = (double)n;
	if (n <= x)
		return(retval);
		
	retval = 1.0;
	for (ctr=(double)n; ctr>(double)x; ctr-=1.0)
		retval *= ctr;
	
	return(retval);
	}
	
	
	
double gamma(n)
	double n;
/*
   gamma() calculates the gamma function.
   
   Entry: n >= 1, is the argument to the gamma function.
   
   Exit:  Returns the gamma function value for the argument n.
          Returns -1.0 if error.
*/
	{
	double	retval;			/* return value */
	double	m;			/* interval for integrating */
	int	idex;			/* index into gamtable array */
	extern double gamtable[];	/* in heliglob.h */
	
	
	if (n < 1.0)
		retval = -1.0;

	else
		{
		m = 1.0;
		while (n >= 2.0)
			{
			m *= n;
			n -= 1.0;
			}
			
/* get fractional part of n, multiply by 100 to find idex */
		idex = (int)(0.5000001 + 100.0 * (n - 1.0));
		retval = m * gamtable[idex];
		}

	return(retval);
	}

